#ifndef MYSTRING__H__
#define MYSTRING__H__

// Copyright (C) 1998 Bruce Guenter <bruce.guenter@qcc.sk.ca>
//
// This package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

#include <string.h>

struct mystringrep
{
  unsigned length;
  unsigned references;
  char buf[1];

  void attach();
  void detach();
  
  static mystringrep* alloc(unsigned);
  static mystringrep* dup(const char*, unsigned);
  static mystringrep* dup(const char*, unsigned,
			  const char*, unsigned);
};

extern mystringrep nil;

class mystring;
class mystringtmp
{
  friend class mystring;
private:
  mystringrep* rep;
  const char* str;
  unsigned len;
  static const unsigned badlen = (unsigned)-1;
  void checkstr()
    {
      // If the string is NULL or the length is zero or
      // the length is unset and the string is empty:
      // set the representation to nil
      if(!str || !len || (len == badlen && !str[0])) {
	rep = &nil;
	str = rep->buf;
	len = 0;
      }
    }
protected:
  unsigned length()
    {
      checkstr();
      if(len == badlen)
	len = strlen(str);
      return len;
    }
  operator const char*()
    {
      checkstr();
      return str;
    }
  bool operator!()
    {
      checkstr();
      return len == 0;
    }
  mystringrep* getrep()
    {
      if(!rep)
	rep = mystringrep::dup(str, length());
      return rep;
    }
public:
  mystringtmp(const char* s) : rep(0), str(s), len(badlen) { }
  mystringtmp(const char* s, unsigned l) : rep(0), str(s), len(l) { }
  mystringtmp(mystringrep* r) : rep(r), str(r->buf), len(r->length) { }
  mystringtmp(const mystring&);
};
      
class mystring
{
  friend class mystringtmp;
private:
  mystringrep* rep;

protected:
  void dupnil();
  void dup(mystringtmp);
  void dup(mystringtmp, mystringtmp);
  void dup(mystringtmp, mystringtmp, mystringtmp);
  void dup(mystringtmp, mystringtmp, mystringtmp, mystringtmp);
  void dup(mystringtmp, mystringtmp, mystringtmp, mystringtmp, mystringtmp);
  void dup(mystringtmp, mystringtmp, mystringtmp,
	   mystringtmp, mystringtmp, mystringtmp);
  void append(mystringtmp);
  void assign(mystringtmp);
public:
  mystring() { dupnil(); }
  mystring(const char* s) { dup(s); }
  mystring(const mystring& s) { dup(s); }
  mystring(const char* str, size_t len) { dup(mystringtmp(str, len)); }
  mystring(mystringtmp a, mystringtmp b) { dup(a, b); }
  mystring(mystringtmp a, mystringtmp b, mystringtmp c) { dup(a, b, c); }
  mystring(mystringtmp a, mystringtmp b, mystringtmp c, mystringtmp d)
    {
      dup(a, b, c, d);
    }
  mystring(mystringtmp a, mystringtmp b, mystringtmp c,
	   mystringtmp d, mystringtmp e)
    {
      dup(a, b, c, d, e);
    }
  mystring(mystringtmp a, mystringtmp b, mystringtmp c,
	   mystringtmp d, mystringtmp e, mystringtmp f)
    {
      dup(a, b, c, d, e, f);
    }
  ~mystring();

  const char* c_str() const { return rep->buf; }

  bool operator!() const { return empty(); }
  
  char operator[](size_t i) const { return rep->buf[i]; }
  
  size_t length() const { return rep->length; }

  bool empty() const { return rep->length == 0; }
  
  int operator!=(mystringtmp in) const
    {
      const char* i = in;
      if(rep->buf == i)
	return 0;
      return strcmp(rep->buf, i);
    }
  bool operator==(mystringtmp in) const
    {
      return !operator!=(in);
    }
  
  mystring operator=(const char* in)
    {
      assign(in);
      return *this;
    }
  mystring operator=(const mystring& in)
    {
      assign(in);
      return *this;
    }
  
  mystring subst(char from, char to) const;
  
  mystring lower() const;
  mystring upper() const;

  int find(char, size_t = 0) const;
  int find_first(mystringtmp, size_t = 0) const;

  int find_last(char) const;
  int find_last(mystringtmp) const;
  
  mystring left(size_t) const;
  mystring right(size_t) const;
  mystring sub(size_t, size_t) const;

  mystring lstrip() const;
  mystring rstrip() const;
  mystring strip() const;

  mystring operator+=(mystringtmp that)
  {
    append(that);
    return *this;
  }
  
  mystring operator+=(char ch)
  {
    char str[2] = { ch, 0 };
    append(mystringtmp(str, 1));
    return *this;
  }
};

inline mystringtmp::mystringtmp(const mystring& s)
  : rep(s.rep), str(s.rep->buf), len(s.rep->length)
{
}
			
mystring operator+(mystringtmp, mystringtmp);

typedef mystring string;

#endif
